from smc.robots.interfaces.mobile_base_interface import MobileBaseInterface
from smc.robots.interfaces.single_arm_interface import SingleArmInterface
from smc.robots.interfaces.whole_body_interface import SingleArmWholeBodyInterface, DualArmWholeBodyInterface
from smc.robots.interfaces.mobile_base_interface import MobileBaseInterface
from smc.control.control_loop_manager import ControlLoopManager
from smc.multiprocessing.process_manager import ProcessManager
from smc.control.optimal_control.abstract_croco_ocp import CrocoOCP
from smc.control.optimal_control.path_following_croco_ocp import (
    BasePathFollowingOCP,
    CrocoEEPathFollowingOCP,
    BaseAndEEPathFollowingOCP,
    BaseAndDualArmEEPathFollowingOCP,
)
from smc.path_generation.path_math.path2d_to_6d import (
    path2D_to_SE3,
)
from smc.path_generation.path_math.cart_pulling_path_math import (
    construct_EE_path,
)
from smc.path_generation.path_math.path_to_trajectory import path2D_to_trajectory2D
from smc.control.controller_templates.path_following_template import (
    PathFollowingFromPlannerControlLoop,
)

import numpy as np
from functools import partial
import types
from argparse import Namespace
from pinocchio import SE3, log6
from collections import deque


def CrocoBasePathFollowingFromPlannerMPCControlLoop(
    ocp: CrocoOCP,
    path2D_untimed: np.ndarray,
    args: Namespace,
    robot: MobileBaseInterface,
    t: int,
    _: dict[str, deque[np.ndarray]],
) -> tuple[np.ndarray, dict[str, np.ndarray], dict[str, np.ndarray]]:

    p = robot.T_w_b.translation[:2]
    max_base_v = np.linalg.norm(robot._max_v[:2])
    path_base = path2D_to_trajectory2D(args, path2D_untimed, max_base_v)
    path_base = np.hstack((path_base, np.zeros((len(path_base), 1))))

    if args.visualizer:
        if t % int(np.ceil(args.ctrl_freq / 25)) == 0:
            robot.visualizer_manager.sendCommand({"path": path_base})

    x0 = np.concatenate([robot.q, robot.v])
    ocp.warmstartAndReSolve(x0, data=(path_base))
    xs = np.array(ocp.solver.xs)
    v_cmd = xs[1, robot.model.nq :]

    err_vector_base = np.linalg.norm(p - path_base[0][:2])  # z axis is irrelevant
    log_item = {}
    log_item["err_norm_base"] = np.linalg.norm(err_vector_base).reshape((1,))
    return v_cmd, {}, log_item


def CrocoBasePathFollowingMPC(
    args: Namespace,
    robot: MobileBaseInterface,
    x0: np.ndarray,
    path_planner: ProcessManager | types.FunctionType,
) -> None:
    """
    CrocoBasePathFollowingMPC
    -----
    """

    ocp = BasePathFollowingOCP(args, robot, x0)
    x0 = np.concatenate([robot.q, robot.v])
    ocp.solveInitialOCP(x0)

    if type(path_planner) == types.FunctionType:
        raise NotImplementedError
    else:
        get_position = lambda robot: robot.T_w_b.translation[:2]
        controlLoop = partial(
            PathFollowingFromPlannerControlLoop,
            path_planner,
            get_position,
            ocp,
            CrocoBasePathFollowingFromPlannerMPCControlLoop,
            args,
            robot,
        )
    log_item = {
        "qs": np.zeros(robot.nq),
        "dqs": np.zeros(robot.nv),
        "err_norm_base": np.zeros((1,)),
    }
    save_past_item = {}
    loop_manager = ControlLoopManager(
        robot, controlLoop, args, save_past_item, log_item
    )
    loop_manager.run()


def CrocoEEPathFollowingMPCControlLoop(
    args: Namespace,
    robot: SingleArmInterface,
    ocp: CrocoOCP,
    path_planner: types.FunctionType,
    t: int,
    _: dict[str, deque[np.ndarray]],
) -> tuple[np.ndarray, dict[str, np.ndarray],dict[str, np.ndarray]]:
    """
    CrocoPathFollowingMPCControlLoop
    -----------------------------
    end-effector(s) follow their path(s).

    path planner can either be a function which spits out a list of path points
    or an instance of ProcessManager which spits out path points
    by calling ProcessManager.getData()
    """
    breakFlag = False
    log_item = {}
    save_past_item = {}

    q = robot.q
    T_w_e = robot.T_w_e

    path_EE_SE3 = path_planner(T_w_e, t)

    x0 = np.concatenate([robot.q, robot.v])
    ocp.warmstartAndReSolve(x0, data=path_EE_SE3)
    xs = np.array(ocp.solver.xs)
    us = np.array(ocp.solver.us)
    vel_cmds = xs[1, robot.model.nq :]

    robot.sendVelocityCommand(vel_cmds)

    # TODO: EVIL AND HAS TO BE REMOVED FROM HERE
    if args.visualizer:
        if t % int(np.ceil(args.ctrl_freq / 25)) == 0:
            robot.visualizer_manager.sendCommand({"frame_path": path_EE_SE3})

    err_vector_ee = log6(robot.T_w_e.actInv(path_EE_SE3[0]))
    log_item["err_vec_ee"] = err_vector_ee
    log_item["qs"] = q.reshape((robot.model.nq,))
    log_item["dqs"] = robot.v.reshape((robot.model.nv,))
    return breakFlag, save_past_item, log_item


def CrocoEEPathFollowingFromPlannerMPCControlLoop(
    ocp: CrocoOCP,
    path2D_untimed: np.ndarray,
    args: Namespace,
    robot: SingleArmInterface,
    t: int,
    _: dict[str, deque[np.ndarray]],
) -> tuple[np.ndarray, dict[str, np.ndarray], dict[str, np.ndarray]]:
    """
    CrocoPathFollowingMPCControlLoop
    -----------------------------
    end-effector(s) follow their path(s).

    path planner can either be a function which spits out a list of path points
    or an instance of ProcessManager which spits out path points
    by calling ProcessManager.getData()
    """
    max_base_v = np.linalg.norm(robot._max_v[:2])
    perc_of_max_v = 0.5
    velocity = perc_of_max_v * max_base_v
    trajectory2D = path2D_to_trajectory2D(args, path2D_untimed, velocity)

    # create a 3D reference out of the path
    path_EE_SE3 = path2D_to_SE3(trajectory2D, args.handlebar_height)
    #    for pose in path_EE_SE3:
    #        print(pose.translation)

    # TODO: EVIL AND HAS TO BE REMOVED FROM HERE
    if args.visualizer:
        if t % int(np.ceil(args.ctrl_freq / 25)) == 0:
            robot.visualizer_manager.sendCommand({"frame_path": path_EE_SE3})

    x0 = np.concatenate([robot.q, robot.v])
    ocp.warmstartAndReSolve(x0, data=path_EE_SE3)
    xs = np.array(ocp.solver.xs)
    v_cmd = xs[1, robot.model.nq :]

    err_vector_ee = log6(robot.T_w_e.actInv(path_EE_SE3[0])).vector
    log_item = {"err_vec_ee": err_vector_ee}

    return v_cmd, {}, log_item


def CrocoEEPathFollowingMPC(
    args: Namespace,
    robot: SingleArmInterface,
    x0: np.ndarray,
    path_planner: ProcessManager | types.FunctionType,
) -> None:
    """
    CrocoEndEffectorPathFollowingMPC
    -----
    follow a fixed pre-determined path, or a path received from a planner.
    the path does NOT need to start from your current pose - you need to get to it yourself.
    """

    ocp = CrocoEEPathFollowingOCP(args, robot, x0)
    # technically should be done in controlloop because now
    # it's solved 2 times before the first command,
    # but we don't have time for details rn
    x0 = np.concatenate([robot.q, robot.v])
    ocp.solveInitialOCP(x0)

    if type(path_planner) == types.FunctionType:
        controlLoop = partial(
            CrocoEEPathFollowingMPCControlLoop, args, robot, ocp, path_planner
        )
    else:
        get_position = lambda robot: robot.T_w_e.translation[:2]
        controlLoop = partial(
            PathFollowingFromPlannerControlLoop,
            path_planner,
            get_position,
            ocp,
            CrocoEEPathFollowingFromPlannerMPCControlLoop,
            args,
            robot,
        )
    log_item = {
        "qs": np.zeros(robot.model.nq),
        "dqs": np.zeros(robot.model.nv),
        "err_vec_ee": np.zeros(6),
    }
    save_past_item = {}
    loop_manager = ControlLoopManager(
        robot, controlLoop, args, save_past_item, log_item
    )
    loop_manager.run()


def BaseAndEEPathFollowingMPCControlLoop(
    args: Namespace,
    robot,
    ocp: CrocoOCP,
    path_planner: types.FunctionType,
    t: int,
    _: dict[str, deque[np.ndarray]],
) -> tuple[np.ndarray, dict[str, np.ndarray]]:
    """
    CrocoPathFollowingMPCControlLoop
    -----------------------------
    end-effector(s) follow their path(s).

    path planner can either be a function which spits out a list of path points
    or an instance of ProcessManager which spits out path points
    by calling ProcessManager.getData()
    """
    breakFlag = False
    log_item = {}
    save_past_dict = {}

    q = robot.q
    T_w_e = robot.T_w_e
    path_base, path_EE = path_planner(T_w_e, t)

    x0 = np.concatenate([robot.q, robot.v])
    ocp.warmstartAndReSolve(x0, data=(path_base, path_EE))
    xs = np.array(ocp.solver.xs)
    us = np.array(ocp.solver.us)
    vel_cmds = xs[1, robot.model.nq :]

    robot.sendVelocityCommand(vel_cmds)
    if t % int(np.ceil(args.ctrl_freq / 25)) == 0:
        path_base = np.array(path_base)
        robot.visualizer_manager.sendCommand({"path": path_base})
        robot.visualizer_manager.sendCommand({"frame_path": path_EE})

    err_vector_ee = log6(T_w_e.actInv(path_EE[0]))
    err_vector_base = np.linalg.norm(q[:2] - path_base[0][:2])  # z axis is irrelevant
    log_item["err_vec_ee"] = err_vector_ee
    log_item["err_norm_ee"] = np.linalg.norm(err_vector_ee).reshape((1,))
    log_item["err_norm_base"] = np.linalg.norm(err_vector_base).reshape((1,))
    log_item["qs"] = q.reshape((robot.model.nq,))
    log_item["dqs"] = robot.v.reshape((robot.model.nv,))
    return breakFlag, save_past_dict, log_item


def BaseAndEEPathFollowingFromPlannerMPCControlLoop(
    ocp: CrocoOCP,
    path2D_untimed_base: np.ndarray,
    args: Namespace,
    robot: SingleArmWholeBodyInterface,
    t: int,
    past_data: dict[str, deque[np.ndarray]],
) -> tuple[np.ndarray, dict[str, np.ndarray], dict[str, np.ndarray]]:
    p = robot.q[:2]

    # NOTE: this one is obtained as the future path from path planner
    max_base_v = np.linalg.norm(robot._max_v[:2])
    trajectory_base = path2D_to_trajectory2D(args, path2D_untimed_base, max_base_v)
    trajectory_base = np.hstack((trajectory_base, np.zeros((len(trajectory_base), 1))))

    trajectorySE3_handlebar = construct_EE_path(args, p, past_data["path2D_untimed"])

    if args.visualizer:
        if t % int(np.ceil(args.ctrl_freq / 25)) == 0:
            robot.visualizer_manager.sendCommand({"path": trajectory_base})
            robot.visualizer_manager.sendCommand({"frame_path": trajectorySE3_handlebar})

    x0 = np.concatenate([robot.q, robot.v])
    ocp.warmstartAndReSolve(x0, data=(trajectory_base, trajectorySE3_handlebar))
    xs = np.array(ocp.solver.xs)
    v_cmd = xs[1, robot.model.nq :]

    err_vector_ee = log6(robot.T_w_e.actInv(trajectorySE3_handlebar[0]))
    err_vector_base = np.linalg.norm(p - trajectory_base[0][:2])  # z axis is irrelevant
    log_item = {}
    log_item["err_vec_ee"] = err_vector_ee
    log_item["err_norm_ee"] = np.linalg.norm(err_vector_ee).reshape((1,))
    log_item["err_norm_base"] = np.linalg.norm(err_vector_base).reshape((1,))
    save_past_item = {"path2D_untimed": p}
    return v_cmd, save_past_item, log_item


def initializePastData(
    args: Namespace, T_w_e: SE3, p_base: np.ndarray, max_base_v: float
) -> np.ndarray:
    # prepopulate past data to make base and cart be on the same path in the past
    # (which didn't actually happen because this just started)
    p_ee = T_w_e.translation[:2]
    straight_line_path = np.linspace(p_ee, p_base, args.past_window_size)
    # straight_line_path_timed = path2D_timed(args, straight_line_path, max_base_v)
    # return straight_line_path_timed # this one is shortened to args.n_knots! and we want the whole buffer
    return straight_line_path


def BaseAndEEPathFollowingMPC(
    args: Namespace,
    robot: SingleArmWholeBodyInterface,
    path_planner: ProcessManager | types.FunctionType,
) -> None:
    """
    BaseAndEEPathFollowingMPC
    -----
    run mpc for a point-to-point inverse kinematics.
    note that the actual problem is solved on
    a dynamics level, and velocities we command
    are actually extracted from the state x(q,dq).
    """

    robot._mode = SingleArmWholeBodyInterface.control_mode.whole_body
    T_w_e = robot.T_w_e
    x0 = np.concatenate([robot.q, robot.v])
    ocp = BaseAndEEPathFollowingOCP(args, robot, x0)
    ocp.solveInitialOCP(x0)

    max_base_v = np.linalg.norm(robot._max_v[:2])

    path2D_handlebar = initializePastData(args, T_w_e, robot.q[:2], float(max_base_v))

    if type(path_planner) == types.FunctionType:
        controlLoop = partial(
            BaseAndEEPathFollowingMPCControlLoop, args, robot, ocp, path_planner
        )
    else:
        get_position = lambda robot: robot.q[:2]
        controlLoop = partial(
            PathFollowingFromPlannerControlLoop,
            path_planner,
            get_position,
            ocp,
            BaseAndEEPathFollowingFromPlannerMPCControlLoop,
            args,
            robot,
        )
    log_item = {
        "qs": np.zeros(robot.model.nq),
        "dqs": np.zeros(robot.model.nv),
        "err_vec_ee": np.zeros((6,)),
        "err_norm_ee": np.zeros((1,)),
        "err_norm_base": np.zeros((1,)),
    }
    save_past_dict = {"path2D_untimed": T_w_e.translation[:2]}
    loop_manager = ControlLoopManager(
        robot, controlLoop, args, save_past_dict, log_item
    )

    # actually put past data into the past window
    loop_manager.past_data["path2D_untimed"].clear()
    loop_manager.past_data["path2D_untimed"].extend(
        path2D_handlebar[i] for i in range(args.past_window_size)
    )

    loop_manager.run()


# TODO: the robot put in is a whole body robot,
# which you need to implement
def BaseAndDualEEPathFollowingMPCControlLoop(
    T_absgoal_l: SE3,
    T_absgoal_r: SE3,
    ocp: CrocoOCP,
    path2D_untimed_base: np.ndarray,
    args: Namespace,
    robot: DualArmWholeBodyInterface,
    t: int,
    past_data: dict[str, deque[np.ndarray]],
) -> tuple[np.ndarray, dict[str, np.ndarray], dict[str, np.ndarray]]:
    """
    BaseAndDualEEPathFollowingMPCControlLoop
    -----------------------------
    cart pulling dual arm control loop
    """
    p = robot.q[:2]

    # NOTE: this one is obtained as the future path from path planner
    max_base_v = np.linalg.norm(robot._max_v[:2])
    trajectory_base = path2D_to_trajectory2D(args, path2D_untimed_base, max_base_v)
    trajectory_base = np.hstack((trajectory_base, np.zeros((len(trajectory_base), 1))))

    trajectorySE3_handlebar = construct_EE_path(args, p, past_data["path2D_untimed"])
    trajectorySE3_l = []
    trajectorySE3_r = []
    for traj_pose in trajectorySE3_handlebar:
        trajectorySE3_l.append(T_absgoal_l.act(traj_pose))
        trajectorySE3_r.append(T_absgoal_r.act(traj_pose))

    if args.visualizer:
        if t % int(np.ceil(args.ctrl_freq / 25)) == 0:
            robot.visualizer_manager.sendCommand({"path": trajectory_base})
            robot.visualizer_manager.sendCommand({"frame_path": trajectorySE3_handlebar})

    x0 = np.concatenate([robot.q, robot.v])
    ocp.warmstartAndReSolve(x0, data=(trajectory_base, trajectorySE3_l, trajectorySE3_r))
    xs = np.array(ocp.solver.xs)
    v_cmd = xs[1, robot.model.nq :]

    err_vector_ee_l = log6(robot.T_w_l.actInv(trajectorySE3_l[0]))
    err_norm_ee_l = np.linalg.norm(err_vector_ee_l)
    err_vector_ee_r = log6(robot.T_w_l.actInv(trajectorySE3_r[0]))
    err_norm_ee_r = np.linalg.norm(err_vector_ee_r)
    err_vector_base = np.linalg.norm(p - trajectory_base[0][:2])  # z axis is irrelevant
    log_item = {}
    log_item["err_vec_ee_l"] = err_vector_ee_l
    log_item["err_norm_ee_l"] = err_norm_ee_l.reshape((1,))
    log_item["err_vec_ee_r"] = err_vector_ee_r
    log_item["err_norm_ee_r"] = err_norm_ee_r.reshape((1,))
    log_item["err_norm_base"] = np.linalg.norm(err_vector_base).reshape((1,))
    save_past_item = {"path2D_untimed": p}
    return v_cmd, save_past_item, log_item

def BaseAndDualEEPathFollowingMPC(
    args: Namespace,
    robot: DualArmWholeBodyInterface,
    path_planner: ProcessManager | types.FunctionType,
    T_absgoal_l: SE3,
    T_absgoal_r: SE3,
    run=True,
) -> None | ControlLoopManager:
    """
    CrocoEndEffectorPathFollowingMPC
    -----
    run mpc for a point-to-point inverse kinematics.
    note that the actual problem is solved on
    a dynamics level, and velocities we command
    are actually extracted from the state x(q,dq).
    """
    robot._mode = DualArmWholeBodyInterface.control_mode.whole_body
    # NOTE: i'm shoving these into the class here - bad style,
    # but i don't 
    T_w_abs = robot.T_w_abs
    x0 = np.concatenate([robot.q, robot.v])
    ocp = BaseAndDualArmEEPathFollowingOCP(args, robot, x0)
    ocp.solveInitialOCP(x0)

    max_base_v = np.linalg.norm(robot._max_v[:2])

    path2D_handlebar = initializePastData(args, T_w_abs, robot.q[:2], float(max_base_v))

    if type(path_planner) == types.FunctionType:
        raise NotImplementedError
    else:
        get_position = lambda robot: robot.q[:2]
        BaseAndDualEEPathFollowingMPCControlLoop_with_l_r = partial(BaseAndDualEEPathFollowingMPCControlLoop, T_absgoal_l, T_absgoal_r)
        controlLoop = partial(
            PathFollowingFromPlannerControlLoop,
            path_planner,
            get_position,
            ocp,
            BaseAndDualEEPathFollowingMPCControlLoop_with_l_r,
            args,
            robot,
        )

    log_item = {
        "qs": np.zeros(robot.model.nq),
        "dqs": np.zeros(robot.model.nv),
        "err_vec_ee_l": np.zeros((6,)),
        "err_norm_ee_l": np.zeros((1,)),
        "err_vec_ee_r": np.zeros((6,)),
        "err_norm_ee_r": np.zeros((1,)),
        "err_norm_base": np.zeros((1,)),
    }
    save_past_dict = {"path2D_untimed": T_w_abs.translation[:2]}
    loop_manager = ControlLoopManager(
        robot, controlLoop, args, save_past_dict, log_item
    )
    # actually put past data into the past window
    loop_manager.past_data["path2D_untimed"].clear()
    loop_manager.past_data["path2D_untimed"].extend(
        path2D_handlebar[i] for i in range(args.past_window_size)
    )

    if run:
        loop_manager.run()
    else:
        return loop_manager
