from smc.control.optimal_control.abstract_croco_ocp import CrocoOCP
from smc.robots.interfaces.mobile_base_interface import MobileBaseInterface
from smc.robots.interfaces.single_arm_interface import SingleArmInterface
from smc.robots.interfaces.dual_arm_interface import DualArmInterface
from smc.robots.interfaces.whole_body_interface import (
    SingleArmWholeBodyInterface,
    DualArmWholeBodyInterface,
)

import numpy as np
import crocoddyl
from argparse import Namespace


class BasePathFollowingOCP(CrocoOCP):
    """
    createBaseAndEEPathFollowingOCP
    -------------------------------
    creates a path following problem.
    it is instantiated to just to stay at the current position.
    NOTE: the path MUST be time indexed with the SAME time used between the knots
    """

    def __init__(self, args, robot: MobileBaseInterface, x0):
        goal = None
        super().__init__(args, robot, x0, goal)

    def constructTaskObjectiveFunction(self, goal) -> None:
        path_base = [np.append(self.x0[:2], 0.0)] * self.args.n_knots

        for i in range(self.args.n_knots):
            baseTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
                self.state, self.robot.base_frame_id, path_base[i], self.state.nv
            )
            baseTrackingCost = crocoddyl.CostModelResidual(
                self.state, baseTranslationResidual
            )
            self.runningCostModels[i].addCost(
                "base_translation" + str(i),
                baseTrackingCost,
                self.args.base_translation_cost,
            )

        self.terminalCostModel.addCost(
            "base_translation" + str(self.args.n_knots),
            baseTrackingCost,
            self.args.base_translation_cost,
        )

    def updateCosts(self, data):
        path_base = data
        for i, runningModel in enumerate(self.solver.problem.runningModels):
            runningModel.differential.costs.costs[
                "base_translation" + str(i)
            ].cost.residual.reference = path_base[i]

        # idk if that's necessary
        self.solver.problem.terminalModel.differential.costs.costs[
            "base_translation" + str(self.args.n_knots)
        ].cost.residual.reference = path_base[-1]


class CrocoEEPathFollowingOCP(CrocoOCP):
    """
    createCrocoEEPathFollowingOCP
    -------------------------------
    creates a path following problem with a single end-effector reference.
    it is instantiated to just to stay at the current position.
    NOTE: the path MUST be time indexed with the SAME time used between the knots
    """

    def __init__(self, args: Namespace, robot: SingleArmInterface, x0: np.ndarray):
        goal = None
        super().__init__(args, robot, x0, goal)

    def constructTaskObjectiveFunction(self, goal):
        T_w_e = self.robot.T_w_e
        path = [T_w_e] * self.args.n_knots

        for i in range(self.args.n_knots):
            framePlacementResidual = crocoddyl.ResidualModelFramePlacement(
                self.state,
                self.robot.ee_frame_id,
                path[i],  # this better be done with the same time steps as the knots
                # NOTE: this should be done outside of this function to clarity
                self.state.nv,
            )
            goalTrackingCost = crocoddyl.CostModelResidual(
                self.state, framePlacementResidual
            )
            self.runningCostModels[i].addCost(
                "gripperPose" + str(i), goalTrackingCost, 1e2
            )
            # runningCostModel.addCost("gripperPose", goalTrackingCost, 1e2)

        self.terminalCostModel.addCost(
            "gripperPose" + str(self.args.n_knots), goalTrackingCost, 1e2
        )
        # terminalCostModel.addCost("gripperPose", goalTrackingCost, 1e2)

    def updateCosts(self, data):
        for i, runningModel in enumerate(self.solver.problem.runningModels):
            runningModel.differential.costs.costs[
                "gripperPose" + str(i)
            ].cost.residual.reference = data[i]

        self.solver.problem.terminalModel.differential.costs.costs[
            "gripperPose" + str(self.args.n_knots)
        ].cost.residual.reference = data[-1]


class BaseAndEEPathFollowingOCP(CrocoOCP):
    """
    createBaseAndEEPathFollowingOCP
    -------------------------------
    creates a path following problem.
    it is instantiated to just to stay at the current position.
    NOTE: the path MUST be time indexed with the SAME time used between the knots
    """

    def __init__(self, args, robot: SingleArmWholeBodyInterface, x0):
        goal = None
        super().__init__(args, robot, x0, goal)

    def constructTaskObjectiveFunction(self, goal) -> None:
        T_w_e = self.robot.T_w_e
        path_ee = [T_w_e] * self.args.n_knots
        path_base = [np.append(self.x0[:2], 0.0)] * self.args.n_knots

        for i in range(self.args.n_knots):
            eePoseResidual = crocoddyl.ResidualModelFramePlacement(
                self.state,
                self.robot.ee_frame_id,
                path_ee[i],  # this better be done with the same time steps as the knots
                # NOTE: this should be done outside of this function to clarity
                self.state.nv,
            )
            eeTrackingCost = crocoddyl.CostModelResidual(self.state, eePoseResidual)
            self.runningCostModels[i].addCost(
                "ee_pose" + str(i), eeTrackingCost, self.args.ee_pose_cost
            )

            baseTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
                self.state, self.robot.base_frame_id, path_base[i], self.state.nv
            )
            baseTrackingCost = crocoddyl.CostModelResidual(
                self.state, baseTranslationResidual
            )
            self.runningCostModels[i].addCost(
                "base_translation" + str(i),
                baseTrackingCost,
                self.args.base_translation_cost,
            )

        self.terminalCostModel.addCost(
            "ee_pose" + str(self.args.n_knots), eeTrackingCost, self.args.ee_pose_cost
        )
        self.terminalCostModel.addCost(
            "base_translation" + str(self.args.n_knots),
            baseTrackingCost,
            self.args.base_translation_cost,
        )

    def updateCosts(self, data):
        path_base = data[0]
        pathSE3 = data[1]
        for i, runningModel in enumerate(self.solver.problem.runningModels):
            runningModel.differential.costs.costs[
                "base_translation" + str(i)
            ].cost.residual.reference = path_base[i]
            runningModel.differential.costs.costs[
                "ee_pose" + str(i)
            ].cost.residual.reference = pathSE3[i]

        # idk if that's necessary
        self.solver.problem.terminalModel.differential.costs.costs[
            "base_translation" + str(self.args.n_knots)
        ].cost.residual.reference = path_base[-1]
        self.solver.problem.terminalModel.differential.costs.costs[
            "ee_pose" + str(self.args.n_knots)
        ].cost.residual.reference = pathSE3[-1]


class DualArmEEPathFollowingOCP(CrocoOCP):
    def __init__(self, args, robot: DualArmInterface, x0):
        goal = None
        raise NotImplementedError
        super().__init__(args, robot, x0, goal)


class BaseAndDualArmEEPathFollowingOCP(CrocoOCP):
    def __init__(self, args, robot: DualArmWholeBodyInterface, x0):
        goal = None
        super().__init__(args, robot, x0, goal)

    def constructTaskObjectiveFunction(self, goal) -> None:
        T_w_e_left, T_w_e_right = self.robot.getLeftRightT_w_e()
        path_ee_left = [T_w_e_left] * self.args.n_knots
        path_ee_right = [T_w_e_right] * self.args.n_knots
        path_base = [np.append(self.x0[:2], 0.0)] * self.args.n_knots

        for i in range(self.args.n_knots):
            l_eePoseResidual = crocoddyl.ResidualModelFramePlacement(
                self.state,
                self.robot.l_ee_frame_id,
                path_ee_left[i],
                self.state.nv,
            )
            l_eeTrackingCost = crocoddyl.CostModelResidual(self.state, l_eePoseResidual)
            self.runningCostModels[i].addCost(
                "l_ee_pose" + str(i), l_eeTrackingCost, self.args.ee_pose_cost
            )
            r_eePoseResidual = crocoddyl.ResidualModelFramePlacement(
                self.state,
                self.robot.r_ee_frame_id,
                path_ee_right[i],
                self.state.nv,
            )
            r_eeTrackingCost = crocoddyl.CostModelResidual(self.state, r_eePoseResidual)
            self.runningCostModels[i].addCost(
                "r_ee_pose" + str(i), r_eeTrackingCost, self.args.ee_pose_cost
            )
            baseTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
                self.state, self.robot.base_frame_id, path_base[i], self.state.nv
            )
            baseTrackingCost = crocoddyl.CostModelResidual(
                self.state, baseTranslationResidual
            )
            self.runningCostModels[i].addCost(
                "base_translation" + str(i),
                baseTrackingCost,
                self.args.base_translation_cost,
            )
        self.terminalCostModel.addCost(
            "l_ee_pose" + str(self.args.n_knots),
            l_eeTrackingCost,
            self.args.ee_pose_cost,
        )
        self.terminalCostModel.addCost(
            "r_ee_pose" + str(self.args.n_knots),
            r_eeTrackingCost,
            self.args.ee_pose_cost,
        )
        self.terminalCostModel.addCost(
            "base_translation" + str(self.args.n_knots),
            baseTrackingCost,
            self.args.base_translation_cost,
        )

    def updateCosts(self, data):
        path_base = data[0]
        pathSE3_l = data[1]
        pathSE3_r = data[2]
        for i, runningModel in enumerate(self.solver.problem.runningModels):
            runningModel.differential.costs.costs[
                "base_translation" + str(i)
            ].cost.residual.reference = path_base[i]
            runningModel.differential.costs.costs[
                "l_ee_pose" + str(i)
            ].cost.residual.reference = pathSE3_l[i]
            runningModel.differential.costs.costs[
                "r_ee_pose" + str(i)
            ].cost.residual.reference = pathSE3_r[i]

        # idk if that's necessary
        self.solver.problem.terminalModel.differential.costs.costs[
            "base_translation" + str(self.args.n_knots)
        ].cost.residual.reference = path_base[-1]
        self.solver.problem.terminalModel.differential.costs.costs[
            "l_ee_pose" + str(self.args.n_knots)
        ].cost.residual.reference = pathSE3_l[-1]
        self.solver.problem.terminalModel.differential.costs.costs[
            "r_ee_pose" + str(self.args.n_knots)
        ].cost.residual.reference = pathSE3_r[-1]
