from smc.control.optimal_control.abstract_croco_ocp import CrocoOCP
from smc.robots.interfaces.single_arm_interface import SingleArmInterface
from smc.robots.interfaces.dual_arm_interface import DualArmInterface
from smc.robots.interfaces.whole_body_interface import (
    SingleArmWholeBodyInterface,
    DualArmWholeBodyInterface,
)

import numpy as np
import pinocchio as pin
import crocoddyl
from argparse import Namespace


class SingleArmIKOCP(CrocoOCP):
    def __init__(
        self,
        args: Namespace,
        robot: SingleArmInterface,
        x0: np.ndarray,
        T_w_goal: pin.SE3,
    ):
        super().__init__(args, robot, x0, T_w_goal)

    def constructTaskObjectiveFunction(self, goal) -> None:
        T_w_goal = goal
        framePlacementResidual = crocoddyl.ResidualModelFramePlacement(
            self.state,
            self.robot.ee_frame_id,
            T_w_goal.copy(),
            self.state.nv,
        )
        goalTrackingCost = crocoddyl.CostModelResidual(
            self.state, framePlacementResidual
        )
        # TODO: final velocity costs only make sense if you're running a single ocp, but not mpc!!
        # TODO: have an argument or something to include or not include them!
        # frameVelocityResidual = crocoddyl.ResidualModelFrameVelocity(
        #    self.state,
        #    self.robot.ee_frame_id,
        #    pin.Motion(np.zeros(6)),
        #    pin.ReferenceFrame.WORLD,
        # )
        # frameVelocityCost = crocoddyl.CostModelResidual(
        #    self.state, frameVelocityResidual
        # )
        for i in range(self.args.n_knots):
            self.runningCostModels[i].addCost(
                "gripperPose" + str(i), goalTrackingCost, 1e2
            )
        self.terminalCostModel.addCost(
            "gripperPose" + str(self.args.n_knots), goalTrackingCost, 1e2
        )
        # self.terminalCostModel.addCost("velFinal", frameVelocityCost, 1e3)

    # there is nothing to update in a point-to-point task
    def updateCosts(self, data):
        pass

    def updateGoalInModels(self, goal) -> None:
        T_w_goal = goal
        for i, runningModel in enumerate(self.solver.problem.runningModels):
            runningModel.differential.costs.costs[
                "gripperPose" + str(i)
            ].cost.residual.reference = T_w_goal
        self.solver.problem.terminalModel.differential.costs.costs[
            "gripperPose" + str(self.args.n_knots)
        ].cost.residual.reference = T_w_goal


class DualArmIKOCP(CrocoOCP):
    def __init__(
        self, args: Namespace, robot: DualArmInterface, x0: np.ndarray, goal: pin.SE3
    ):
        super().__init__(args, robot, x0, goal)

    def constructTaskObjectiveFunction(self, goal) -> None:
        T_w_lgoal, T_w_rgoal = goal
        framePlacementResidual_l = crocoddyl.ResidualModelFramePlacement(
            self.state,
            self.robot.l_ee_frame_id,
            T_w_lgoal.copy(),
            self.state.nv,
        )
        framePlacementResidual_r = crocoddyl.ResidualModelFramePlacement(
            self.state,
            self.robot.r_ee_frame_id,
            T_w_rgoal.copy(),
            self.state.nv,
        )
        goalTrackingCost_l = crocoddyl.CostModelResidual(
            self.state, framePlacementResidual_l
        )
        goalTrackingCost_r = crocoddyl.CostModelResidual(
            self.state, framePlacementResidual_r
        )
        # TODO: final velocity costs only make sense if you're running a single ocp, but not mpc!!
        # TODO: have an argument or something to include or not include them!

        # frameVelocityCost_l = crocoddyl.CostModelResidual(
        #    self.state, frameVelocityResidual_l
        # frameVelocityResidual_l = crocoddyl.ResidualModelFrameVelocity(
        #    self.state,
        #    self.robot.l_ee_frame_id,
        #    pin.Motion(np.zeros(6)),
        #    pin.ReferenceFrame.WORLD,
        # )
        # frameVelocityResidual_r = crocoddyl.ResidualModelFrameVelocity(
        #    self.state,
        #    self.robot.r_ee_frame_id,
        #    pin.Motion(np.zeros(6)),
        #    pin.ReferenceFrame.WORLD,
        # )
        # )
        # frameVelocityCost_r = crocoddyl.CostModelResidual(
        #    self.state, frameVelocityResidual_r
        # )
        for i in range(self.args.n_knots):
            self.runningCostModels[i].addCost(
                "gripperPose_l" + str(i), goalTrackingCost_l, 1e2
            )
            self.runningCostModels[i].addCost(
                "gripperPose_r" + str(i), goalTrackingCost_r, 1e2
            )
        self.terminalCostModel.addCost(
            "gripperPose_l" + str(self.args.n_knots), goalTrackingCost_l, 1e2
        )
        self.terminalCostModel.addCost(
            "gripperPose_r" + str(self.args.n_knots), goalTrackingCost_r, 1e2
        )
        # self.terminalCostModel.addCost("velFinal_l", frameVelocityCost_l, 1e3)
        # self.terminalCostModel.addCost("velFinal_r", frameVelocityCost_r, 1e3)

    # there is nothing to update in a point-to-point task
    def updateCosts(self, data):
        pass

    def updateGoalInModels(self, T_w_lgoal: pin.SE3, T_w_rgoal: pin.SE3) -> None:
        for i, runningModel in enumerate(self.solver.problem.runningModels):
            runningModel.differential.costs.costs[
                "gripperPose_l" + str(i)
            ].cost.residual.reference = T_w_lgoal
            runningModel.differential.costs.costs[
                "gripperPose_r" + str(i)
            ].cost.residual.reference = T_w_rgoal

        self.solver.problem.terminalModel.differential.costs.costs[
            "gripperPose_l" + str(self.args.n_knots)
        ].cost.residual.reference = T_w_lgoal
        self.solver.problem.terminalModel.differential.costs.costs[
            "gripperPose_r" + str(self.args.n_knots)
        ].cost.residual.reference = T_w_rgoal


class BaseAndSingleArmIKOCP(SingleArmIKOCP):
    def __init__(
        self,
        args: Namespace,
        robot: SingleArmWholeBodyInterface,
        x0: np.ndarray,
        goal,
        #        T_w_eegoal: pin.SE3,
        #        p_basegoal: np.ndarray,
    ):
        # T_w_eegoal, p_basegoal = goal
        super().__init__(args, robot, x0, goal)

    def constructTaskObjectiveFunction(
        # self, T_w_eegoal: pin.SE3, p_basegoal: np.ndarray
        self,
        goal,
    ) -> None:
        T_w_eegoal, p_basegoal = goal
        super().constructTaskObjectiveFunction(T_w_eegoal)
        for i in range(self.args.n_knots):
            baseTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
                self.state, self.robot.base_frame_id, p_basegoal, self.state.nv
            )
            baseTrackingCost = crocoddyl.CostModelResidual(
                self.state, baseTranslationResidual
            )
            self.runningCostModels[i].addCost(
                "base_translation" + str(i),
                baseTrackingCost,
                self.args.base_translation_cost,
            )
        self.terminalCostModel.addCost(
            "base_translation" + str(self.args.n_knots),
            baseTrackingCost,
            self.args.base_translation_cost,
        )

    # there is nothing to update in a point-to-point task
    def updateCosts(self, data):
        pass

    def updateGoalInModels(self, goal) -> None:
        # self, T_w_eegoal: pin.SE3, p_basegoal: np.ndarray) -> None:
        T_w_eegoal, p_basegoal = goal
        super().updateGoalInModels(T_w_eegoal)
        for i, runningModel in enumerate(self.solver.problem.runningModels):
            runningModel.differential.costs.costs[
                "base_translation" + str(i)
            ].cost.residual.reference = p_basegoal
        self.solver.problem.terminalModel.differential.costs.costs[
            "base_translation" + str(self.args.n_knots)
        ].cost.residual.reference = p_basegoal


class BaseAndDualArmIKOCP(DualArmIKOCP):
    def __init__(
        self,
        args: Namespace,
        robot: DualArmWholeBodyInterface,
        x0: np.ndarray,
        goal,
    ):
        # T_w_eegoal, p_basegoal = goal
        super().__init__(args, robot, x0, goal)

    def constructTaskObjectiveFunction(
        # self, T_w_eegoal: pin.SE3, p_basegoal: np.ndarray
        self,
        goal,
    ) -> None:
        T_w_lgoal, T_w_rgoal, p_basegoal = goal
        super().constructTaskObjectiveFunction((T_w_lgoal, T_w_rgoal))
        for i in range(self.args.n_knots):
            baseTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
                self.state, self.robot.base_frame_id, p_basegoal, self.state.nv
            )
            baseTrackingCost = crocoddyl.CostModelResidual(
                self.state, baseTranslationResidual
            )
            self.runningCostModels[i].addCost(
                "base_translation" + str(i),
                baseTrackingCost,
                self.args.base_translation_cost,
            )
        self.terminalCostModel.addCost(
            "base_translation" + str(self.args.n_knots),
            baseTrackingCost,
            self.args.base_translation_cost,
        )

    # there is nothing to update in a point-to-point task
    def updateCosts(self, data) -> None:
        pass

    def updateGoalInModels(self, goal) -> None:
        # self, T_w_eegoal: pin.SE3, p_basegoal: np.ndarray) -> None:
        T_w_lgoal, T_w_rgoal, p_basegoal = goal
        super().updateGoalInModels(T_w_lgoal, T_w_rgoal)
        for i, runningModel in enumerate(self.solver.problem.runningModels):
            runningModel.differential.costs.costs[
                "base_translation" + str(i)
            ].cost.residual.reference = p_basegoal
        self.solver.problem.terminalModel.differential.costs.costs[
            "base_translation" + str(self.args.n_knots)
        ].cost.residual.reference = p_basegoal
