from smc.robots.interfaces.mobile_base_interface import MobileBaseInterface
from smc.robots.interfaces.single_arm_interface import SingleArmInterface
from smc.robots.interfaces.dual_arm_interface import DualArmInterface

import pinocchio as pin
from argparse import Namespace
import numpy as np
from enum import Enum

# TODO: put back in when python3.12 will be safe to use
# from typing import override


class SingleArmWholeBodyInterface(SingleArmInterface, MobileBaseInterface):
    """
    SingleArmWholeBodyInterface
    ---------------------------
    exists to provide either:
    1) whole body values
    2) base only value
    3) arm only value

    what you get depends on the mode you set - they're enumerate as above
    """

    control_mode = Enum("mode", [("whole_body", 1), ("base_only", 2), ("arm_only", 3)])

    def __init__(self, args: Namespace):
        if args.debug_prints:
            print("SingleArmWholeBodyInterface init")
        self._mode: SingleArmWholeBodyInterface.control_mode
        super().__init__(args)

    @property
    def mode(self) -> control_mode:
        return self._mode

    @mode.setter
    def setMode(self, mode: control_mode) -> None:
        assert type(mode) is self.control_mode
        self._mode = mode

    # TODO: put back in when python3.12 will be safe to use
    #    @override
    @property
    def q(self) -> np.ndarray:
        if self._mode == self.control_mode.base_only:
            return self._q[:4]

        if self._mode == self.control_mode.arm_only:
            return self._q[4:]

        return self._q.copy()

    @property
    def nq(self):
        if self._mode == self.control_mode.base_only:
            return 4

        if self._mode == self.control_mode.arm_only:
            return self.model.nq - 4

        return self.model.nq

    @property
    def v(self) -> np.ndarray:
        if self._mode == self.control_mode.base_only:
            return self._v[:3]

        if self._mode == self.control_mode.arm_only:
            return self._v[3:]

        return self._v.copy()

    @property
    def nv(self):
        if self._mode == self.control_mode.base_only:
            return 3

        if self._mode == self.control_mode.arm_only:
            return self.model.nv - 3

        return self.model.nv

    # TODO: put back in when python3.12 will be safe to use
    #    @override
    def forwardKinematics(self) -> None:
        pin.forwardKinematics(
            self.model,
            self.data,
            self._q,
        )
        pin.updateFramePlacement(self.model, self.data, self._ee_frame_id)
        pin.updateFramePlacement(self.model, self.data, self._base_frame_id)
        self._T_w_e = self.data.oMf[self._ee_frame_id].copy()
        self._T_w_b = self.data.oMf[self._base_frame_id].copy()

    # TODO: put back in when python3.12 will be safe to use
    #    @override
    def getJacobian(self) -> np.ndarray:
        J_full = pin.computeFrameJacobian(
            self.model, self.data, self._q, self._ee_frame_id
        )
        if self._mode == self.control_mode.base_only:
            return J_full[:, :3]

        if self._mode == self.control_mode.arm_only:
            return J_full[:, 3:]

        return J_full

    # TODO: put back in when python3.12 will be safe to use
    #    @override
    def sendVelocityCommand(self, v) -> None:
        """
        sendVelocityCommand
        -------------------
        1) saturate the command to comply with hardware limits or smaller limits you've set
        2) send it via the particular robot's particular communication interface
        """
        assert type(v) == np.ndarray

        if self._mode == self.control_mode.base_only:
            v = np.hstack((v, np.zeros(self.model.nv - 3)))

        if self._mode == self.control_mode.arm_only:
            v = np.hstack((np.zeros(3), v))

        assert len(v) == self.model.nv
        v = np.clip(v, -1 * self._max_v, self._max_v)
        self.sendVelocityCommandToReal(v)


class DualArmWholeBodyInterface(DualArmInterface, MobileBaseInterface):
    control_mode = Enum(
        "mode",
        [
            ("whole_body", 1),
            ("base_only", 2),
            ("both_arms_only", 3),
            ("left_arm_only", 4),
            ("right_arm_only", 5),
        ],
    )

    def __init__(self, args: Namespace):
        if args.debug_prints:
            print("DualArmWholeBodyInterface init")
        self._mode: DualArmWholeBodyInterface.control_mode
        super().__init__(args)

    # TODO: put back in when python3.12 will be safe to use
    #    @override
    @property
    def q(self) -> np.ndarray:
        if self._mode == self.control_mode.base_only:
            return self._q[:4]

        if self._mode == self.control_mode.both_arms_only:
            return self._q[4:]

        # NOTE: left should be on the left side of the joint values array
        if self._mode == self.control_mode.left_arm_only:
            return self._q[4 : (self.model.nq - 4) // 2]

        # NOTE: right should be on the right side of the joint values array
        if self._mode == self.control_mode.right_arm_only:
            return self._q[4 + (self.model.nq - 4) // 2 :]

        return self._q.copy()

    @property
    def nq(self):
        if self._mode == self.control_mode.base_only:
            return 4

        if self._mode == self.control_mode.both_arms_only:
            return self.model.nq - 4

        if self._mode == self.control_mode.left_arm_only:
            return (self.model.nq - 4) // 2

        if self._mode == self.control_mode.right_arm_only:
            return (self.model.nq - 4) // 2
        return self.model.nq

    @property
    def v(self) -> np.ndarray:
        if self._mode == self.control_mode.base_only:
            return self._v[:3]

        if self._mode == self.control_mode.both_arms_only:
            return self._v[3:]

        if self._mode == self.control_mode.left_arm_only:
            return self._v[3 : (self.model.nv - 3) // 2]

        if self._mode == self.control_mode.right_arm_only:
            return self._v[3 + (self.model.nv - 3) // 2 :]

        return self._v.copy()

    @property
    def nv(self) -> int:
        if self._mode == self.control_mode.base_only:
            return 3

        if self._mode == self.control_mode.both_arms_only:
            return self.model.nv - 3

        if self._mode == self.control_mode.left_arm_only:
            return (self.model.nv - 3) // 2

        if self._mode == self.control_mode.right_arm_only:
            return (self.model.nv - 3) // 2

        return self.model.nv

    # TODO: put back in when python3.12 will be safe to use
    #    @override
    def forwardKinematics(self) -> None:
        pin.forwardKinematics(
            self.model,
            self.data,
            self._q,
        )
        pin.updateFramePlacement(self.model, self.data, self._base_frame_id)
        self._T_w_b = self.data.oMf[self._base_frame_id].copy()
        pin.updateFramePlacement(self.model, self.data, self._l_ee_frame_id)
        pin.updateFramePlacement(self.model, self.data, self._r_ee_frame_id)
        self._T_w_l = self.data.oMf[self._l_ee_frame_id].copy()
        self._T_w_r = self.data.oMf[self._r_ee_frame_id].copy()
        self.T_w_abs  # will update _T_w_abs in the process

    # TODO: put back in when python3.12 will be safe to use
    #    @override
    def getJacobian(self) -> np.ndarray:
        J_left_with_base = pin.computeFrameJacobian(
            self.model, self.data, self._q, self._l_ee_frame_id
        )[:, : 3 + (self.model.nv - 3) // 2]
        if self._mode == self.control_mode.left_arm_only:
            return J_left_with_base[:, 3:]

        # NOTE: the base jacobian can be extracted from either left or right frame -
        # since it's a body jacobian both have to be the same at the base.
        # for efficiency of course it would be best to construct it in place,
        # but who cares if it runs on time either way.
        if self._mode == self.control_mode.base_only:
            return J_left_with_base[:, :3]

        J_right = pin.computeFrameJacobian(
            self.model, self.data, self._q, self._r_ee_frame_id
        )[:, 3 + (self.model.nv - 3) // 2 :]
        if self._mode == self.control_mode.right_arm_only:
            return J_right

        J_full = np.zeros((12, self.nv))
        J_full[:6, : 3 + (self.model.nv - 3) // 2] = J_left_with_base
        J_full[6:, 3 + (self.model.nv - 3) // 2 :] = J_right
        # NOTE: add base for right end-effector
        # look at note above returning base only jacobian
        J_full[6:, :3] = J_left_with_base[:, :3]

        if self._mode == self.control_mode.both_arms_only:
            return J_full[:, 3:]

        return J_full

    # TODO: put back in when python3.12 will be safe to use
    #    @override
    def sendVelocityCommand(self, v) -> None:
        """
        sendVelocityCommand
        -------------------
        1) saturate the command to comply with hardware limits or smaller limits you've set
        2) send it via the particular robot's particular communication interface
        """
        assert type(v) == np.ndarray
        if self._mode == self.control_mode.base_only:
            v = np.hstack((v, np.zeros(self.model.nv - 3)))

        if self._mode == self.control_mode.both_arms_only:
            v = np.hstack((np.zeros(3), v))

        if self._mode == self.control_mode.left_arm_only:
            v = np.zeros(self.model.nv)
            v[3 : (self.model.nv - 3) // 2] = v

        if self._mode == self.control_mode.right_arm_only:
            v = np.zeros(self.model.nv)
            v[3 + (self.model.nv - 3) // 2 :] = v

        assert len(v) == self.model.nv
        v = np.clip(v, -1 * self._max_v, self._max_v)

        self.sendVelocityCommandToReal(v)
